#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import math


class SunPosition:
    def solar_declination(self, day_of_year: int) -> float:
        r"""
        Calculate the solar declination angle in radians.

        The solar declination angle is the angle between the rays of the sun
        and the plane of the Earth's equator.

        The formula used to calculate the solar declination angle is:

        .. math::
            \delta = \frac{23.45 \pi}{180} \times
            \sin \left(\frac{2\pi~(284 + n)}{365}\right)

        | :math:`n` is the day of the year (i.e., ``day_of_year``)

        :param day_of_year: The day of the year.
        :type day_of_year: int
        :return: The solar declination angle in radians.
        :rtype: float
        """
        # tilt of the Earth's axis (in degrees)
        earth_tilt_degrees = 23.45

        # Convert the tilt to radians
        earth_tilt_radians = math.radians(earth_tilt_degrees)

        # Offset to ensure declination angle is zero at the March equinox
        equinox_offset_days = 284

        # Calculate the solar declination angle
        solar_declination = (
            math.sin((2 * math.pi) * (equinox_offset_days + day_of_year) / 365)
            * earth_tilt_radians
        )

        return solar_declination

    def hour_angle(self, solar_time: float) -> float:
        r"""
        Calculate the hour angle based on the solar time.

        The hour angle is a measure of time, expressed in angular terms,
        from solar noon.

        The formula used to calculate the hour angle is:

        .. math::
            \omega = (t - 12\times 60 \times 60)
            \times \frac{\pi}{12\times 60 \times 60}

        :param solar_time: The solar time in seconds.
        :type solar_time: float
        :return: The hour angle in radians.
        :rtype: float
        """
        # The number of seconds in half a day (12 hours)
        seconds_in_half_day = 12 * 60 * 60

        # The Earth rotates by pi/seconds_in_half_day radians per second
        earth_rotation_rate = math.pi / seconds_in_half_day

        # Calculate the hour angle
        hour_angle = (solar_time - seconds_in_half_day) * earth_rotation_rate

        return hour_angle

    def solar_time(self, hour_angle: float) -> float:
        r"""
        Calculate the solar time based on the hour angle.

        The solar time is a measure of time, expressed in seconds, from solar noon.

        The formula used to calculate the solar time is:

        .. math::
            t = \omega \times \frac{12 \times 60 \times 60}{\pi}
            + 12 \times 60 \times 60

        :param hour_angle: The hour angle in radians.
        :type hour_angle: float
        :return: The solar time in seconds.
        :rtype: float
        """
        # The number of seconds in half a day (12 hours)
        seconds_in_half_day = 12 * 60 * 60

        # The Earth rotates by pi/seconds_in_half_day radians per second
        earth_rotation_rate = math.pi / seconds_in_half_day

        # Calculate the solar time
        solar_time = hour_angle / earth_rotation_rate + seconds_in_half_day

        return solar_time



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import csv
import logging
from pathlib import Path

from .logger import logger_decorator
from .numerical_integration import IrradiationCalculator


class ReportGenerator:
    @logger_decorator
    def generate_optimal_orientation_csv_report(
        self,
        path: Path,
        irradiation_calculator: IrradiationCalculator,
        from_day: int,
        to_day: int,
    ) -> None:
        r"""
        This method generates a report of
        optimal solar panel orientation in CSV format.

        :param path: A Path object that points to the CSV file
                     where the report will be written.
        :type path: Path
        :param irradiation_calculator: An instance of the IrradiationCalculator class.
        :type irradiation_calculator: pysolorie.IrradiationCalculator
        :param from_day: The starting day of the report.
        :type from_day: int
        :param to_day: The ending day of the report.
        :type to_day: int
        """
        with open(path, "w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(["Day", "Beta (degrees)"])

            for day in range(from_day, to_day):
                beta = irradiation_calculator.find_optimal_orientation(day)
                logger = logging.getLogger(
                    self.generate_optimal_orientation_csv_report.__name__
                )
                logger.info(
                    f"On day {day},"
                    + f"the solar panel's optimal orientation is {beta} degrees."
                )

                # Write the result to the CSV file
                writer.writerow([day, beta])



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math

from .sun_position import SunPosition


class SolarIrradiance:
    def __init__(self, sun_position: SunPosition):
        r"""
        To instantiate the ``SolarIrradiance`` class, provide the following parameter.


        :param sun_position: An instance of the SunPosition class.
        :type sun_position: pysolorie.SunPosition
        """
        self.sun_position = sun_position

    def calculate_extraterrestrial_irradiance(self, day_of_year: int) -> float:
        r"""
        Calculate the extraterrestrial solar irradiance for a given day of the year.

        The extraterrestrial solar irradiance, :math:`E`,
        is the amount of solar energy received per unit area on
        a surface perpendicular to the Sun's rays outside Earth's atmosphere.

        The formula used is:

        .. math::
            E = SC
            \times (1 + 0.33 \times \cos (\frac{2\pi~n}{365}))


        | - :math:`SC` is the average solar radiation arriving outside
           of the Earth's atmosphere,
           which is approximately ``1367`` Watts per square meter.
           This is also known as the solar constant.
        | - The factor ``0.033`` accounts
          for the variation in the Earth-Sun distance
          due to the Earth's elliptical orbit.
        | - :math:`n` is the day of the year (i.e., ``day_of_year``)

        :param day\_of\_year: The day of the year, ranging from 1 to 365.
        :type day\_of\_year: int

        :return: The extraterrestrial solar irradiance in Watts per square meter.
        :rtype: float
        """

        # Solar constant (W/m^2)
        SOLAR_CONSTANT = 1367

        # Factor to account for the Earth's orbital eccentricity
        earth_orbital_eccentricity = 0.033

        # Calculate the extraterrestrial solar irradiance
        extraterrestrial_irradiance = SOLAR_CONSTANT * (
            1 + earth_orbital_eccentricity * math.cos(2 * math.pi * day_of_year / 365)
        )

        return extraterrestrial_irradiance



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import csv
import math
from pathlib import Path
from typing import Tuple

import matplotlib.pyplot as plt
import pytest

from pysolorie import (
    AtmosphericTransmission,
    HottelModel,
    IrradiationCalculator,
    Observer,
    Plotter,
    ReportGenerator,
    SolarIrradiance,
    SunPosition,
)


@pytest.mark.parametrize(
    "climate_type, observer_altitude, expected_result",
    [
        ("MIDLATITUDE SUMMER", 1200, (0.228, 0.666, 0.309)),  # Tehran Summer
        ("MIDLATITUDE WINTER", 1200, (0.242, 0.679, 0.303)),  # Tehran Winter
        ("TROPICAL", 26, (0.124, 0.739, 0.392)),  # Medan
        ("SUBARCTIC SUMMER", 136, (0.140, 0.739, 0.379)),  # Fairbanks
    ],
)
def test_calculate_transmittance_components(
    climate_type: str,
    observer_altitude: int,
    expected_result: Tuple[float, float, float],
) -> None:
    hottel_model: HottelModel = HottelModel()
    result: Tuple[
        float, float, float
    ] = hottel_model.calculate_transmittance_components(climate_type, observer_altitude)
    assert pytest.approx(result, abs=1e-3) == expected_result


def test_invalid_climate_type() -> None:
    with pytest.raises(ValueError, match="Invalid climate type"):
        hottel_model: HottelModel = HottelModel()
        hottel_model.calculate_transmittance_components("INVALID", 1000)


@pytest.mark.parametrize(
    "day_of_year, solar_time, expected_declination, expected_hour_angle",
    [
        (1, 12 * 60 * 60, -0.4014257279586958, 0),  # January 1st at noon
        (81, 10 * 60 * 60, 0, -math.pi / 6),  # March 22nd at 10am (equinox)
        (81, 12 * 60 * 60, 0, 0),  # March 22nd at noon (equinox)
        (1, 13 * 60 * 60, -0.4014257279586958, math.pi / 12),  # January 1st at 1pm
    ],
)
def test_sun_position(
    day_of_year: int,
    solar_time: int,
    expected_declination: float,
    expected_hour_angle: float,
) -> None:
    sun_position: SunPosition = SunPosition()
    declination: float = sun_position.solar_declination(day_of_year)
    hour_angle: float = sun_position.hour_angle(solar_time)
    assert declination == pytest.approx(expected_declination, abs=1e-3)
    assert hour_angle == pytest.approx(expected_hour_angle, abs=1e-3)


@pytest.mark.parametrize(
    "hour_angle, expected_solar_time",
    [
        (0, 12 * 60 * 60),  # solar noon
        (-math.pi / 6, 10 * 60 * 60),  # 10am
        (math.pi / 12, 13 * 60 * 60),  # 1pm
    ],
)
def test_solar_time(
    hour_angle: float,
    expected_solar_time: int,
) -> None:
    sun_position: SunPosition = SunPosition()
    solar_time: float = sun_position.solar_time(hour_angle)
    assert solar_time == pytest.approx(expected_solar_time, abs=1e-3)


@pytest.mark.parametrize(
    "day_of_year, expected_irradiance",
    [
        (1, 1412.104),  # January 1st
        (81, 1374.918),  # March 22nd (equinox)
        (172, 1322.623),  # June 21st (summer solstice)
        (264, 1359.464),  # September 23rd (equinox)
        (355, 1411.444),  # December 21st (winter solstice)
    ],
)
def test_calculate_extraterrestrial_irradiance(
    day_of_year: int, expected_irradiance: float
) -> None:
    sun_position = SunPosition()
    solar_irradiance = SolarIrradiance(sun_position)
    irradiance: float = solar_irradiance.calculate_extraterrestrial_irradiance(
        day_of_year
    )
    assert irradiance == pytest.approx(expected_irradiance, abs=1e-3)


@pytest.mark.parametrize(
    "observer_latitude,"
    + "observer_longitude,"
    + "day_of_year,"
    + "solar_time,"
    + "expected_zenith_angle",
    [
        (
            35.69,
            51.39,
            81,
            12 * 60 * 60,
            0.623,
        ),  # Test at Tehran on the 81st day of the year at noon
        (
            35.69,
            51.39,
            355,
            12 * 60 * 60,
            1.032,
        ),  # Test at Tehran on the 355th day of the year at noon
        (
            3.59,
            98.67,
            81,
            10 * 60 * 60,
            0.527,
        ),  # Test at Medan on the 81st day of the year at 10am
        (
            64.84,
            -147.72,
            1,
            13 * 60 * 60,
            1.547,
        ),  # Test at Fairbanks on the 1st day of the year at 1pm
        (90, 0, 1, 13 * 60 * 60, 1.972),  # Test at North Pole on January 1st at 1pm
    ],
)
def test_calculate_zenith_angle(
    observer_latitude: float,
    observer_longitude: float,
    day_of_year: int,
    solar_time: int,
    expected_zenith_angle: float,
) -> None:
    observer: Observer = Observer(observer_latitude, observer_longitude)
    zenith_angle: float = observer.calculate_zenith_angle(day_of_year, solar_time)
    assert zenith_angle == pytest.approx(expected_zenith_angle, abs=1e-3)


def test_calculate_zenith_angle_without_latitude():
    observer = Observer(None, 0)
    with pytest.raises(
        ValueError,
        match="Observer latitude must be provided.",
    ):
        observer.calculate_zenith_angle(1, 12 * 60 * 60)


@pytest.mark.parametrize(
    "climate_type,"
    + "observer_altitude,"
    + "observer_latitude,"
    + "day_of_year,"
    + "solar_time,"
    + "expected_transmittance",
    [
        ("MIDLATITUDE SUMMER", 1200, 35.69, 81, 12 * 60 * 60, 0.683),  # Tehran Summer
        ("MIDLATITUDE WINTER", 1200, 35.69, 355, 12 * 60 * 60, 0.618),  # Tehran Winter
        ("TROPICAL", 63, 3.59, 81, 10 * 60 * 60, 0.597),  # Medan
        ("SUBARCTIC SUMMER", 136, 64.84, 1, 13 * 60 * 60, 0.140),  # Fairbanks
    ],
)
def test_calculate_transmittance(
    climate_type: str,
    observer_altitude: int,
    observer_latitude: float,
    day_of_year: int,
    solar_time: float,
    expected_transmittance: float,
) -> None:
    atmospheric_transmission: AtmosphericTransmission = AtmosphericTransmission(
        climate_type, observer_altitude, observer_latitude
    )
    result: float = atmospheric_transmission.calculate_transmittance(
        day_of_year, solar_time
    )
    assert pytest.approx(result, abs=1e-3) == expected_transmittance


@pytest.mark.parametrize(
    "day_of_year, expected_sunrise_hour_angle, expected_sunset_hour_angle",
    [
        (1, -1.261, 1.261),  # January 1st
        (81, -math.pi / 2, math.pi / 2),  # March 22nd (equinox)
        (172, -1.888, 1.888),  # June 21st (solstice)
    ],
)
def test_calculate_sunrise_sunset(
    day_of_year: int,
    expected_sunrise_hour_angle: float,
    expected_sunset_hour_angle: float,
) -> None:
    observer: Observer = Observer(observer_latitude=35.69)  # Tehran
    sunrise_hour_angle, sunset_hour_angle = observer.calculate_sunrise_sunset(
        day_of_year
    )
    assert sunrise_hour_angle == pytest.approx(expected_sunrise_hour_angle, abs=1e-3)
    assert sunset_hour_angle == pytest.approx(expected_sunset_hour_angle, abs=1e-3)


@pytest.mark.parametrize(
    "climate_type, observer_altitude, observer_latitude, day_of_year, expected_result",
    [
        (
            "MIDLATITUDE SUMMER",
            1200,
            35.6892,
            172,
            0.241,
        ),  # Tehran Summer, day_of_year=172 (June 21)
        (
            "MIDLATITUDE WINTER",
            1200,
            35.6892,
            355,
            63.839,
        ),  # Tehran Winter, day_of_year=355 (Dec 21)
        (
            "TROPICAL",
            26,
            3.5952,
            100,
            -6.635,
        ),  # Medan, day_of_year=100 (April 10)
        (
            "SUBARCTIC SUMMER",
            132,
            64.84361,
            200,
            32.613,
        ),  # Fairbanks Summer, day_of_year=200 (July 19)
    ],
)
def test_find_optimal_orientation(
    climate_type: str,
    observer_altitude: int,
    observer_latitude: float,
    day_of_year: int,
    expected_result: float,
) -> None:
    irradiation_calculator = IrradiationCalculator(
        climate_type, observer_altitude, observer_latitude
    )
    result = irradiation_calculator.find_optimal_orientation(day_of_year)
    assert pytest.approx(result, abs=1e-3) == expected_result


def test_generate_optimal_orientation_csv_report(tmpdir):
    # Create a temporary directory for the test
    temp_dir = Path(tmpdir)

    # Initialize the ReportGenerator
    report_generator = ReportGenerator()

    # Initialize the IrradiationCalculator for Tehran
    irradiation_calculator = IrradiationCalculator("MIDLATITUDE SUMMER", 1200, 35.6892)

    # Define the path for the CSV file
    csv_path = temp_dir / "report.csv"
    from_day = 60
    to_day = 70
    # Call the method to generate the report
    report_generator.generate_optimal_orientation_csv_report(
        csv_path, irradiation_calculator, from_day, to_day
    )

    # Check the CSV file
    with open(csv_path, "r") as file:
        reader = csv.reader(file)
        header = next(reader)
        assert header == ["Day", "Beta (degrees)"]
        for i, row in enumerate(reader, start=from_day):
            day, beta = row
            assert int(day) == i
            expected_beta = irradiation_calculator.find_optimal_orientation(i)
            assert pytest.approx(float(beta), abs=1e-3) == expected_beta


def test_plot_optimal_orientation(tmpdir):
    # Create a temporary directory for the test
    temp_dir = Path(tmpdir)

    # Initialize the Plotter
    plotter = Plotter()

    # Initialize the IrradiationCalculator for Tehran
    irradiation_calculator = IrradiationCalculator("MIDLATITUDE SUMMER", 1200, 35.6892)

    # Define the path for the plot
    plot_path = temp_dir / "plot.png"
    from_day = 60
    to_day = 70
    # Call the method to generate the plot
    plotter.plot_optimal_orientation(
        irradiation_calculator, from_day, to_day, plot_path
    )

    # Check the plot file
    assert plot_path.exists(), "The plot file was not created."

    img = plt.imread(plot_path)
    assert img.shape[0] > 0, "The plot image has no content."
    assert img.shape[1] > 0, "The plot image has no content."



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Dict, Tuple


class HottelModel:
    r"""
    Hottel Model for estimating clear-sky beam radiation transmittance
    based on climate type, and observer altitude.

    ``CLIMATE_CONSTANTS``: Correction factors for different climate types
    :math:`r_0`, :math:`r_1`, and :math:`r_k`.
    """

    CLIMATE_CONSTANTS: Dict[str, Tuple[float, float, float]] = {
        "TROPICAL": (0.95, 0.98, 1.02),
        "MIDLATITUDE SUMMER": (0.97, 0.99, 1.02),
        "SUBARCTIC SUMMER": (0.99, 0.99, 1.01),
        "MIDLATITUDE WINTER": (1.03, 1.01, 1.00),
    }

    def _convert_to_km(self, observer_altitude: int) -> float:
        r"""
        Convert altitude from meters to kilometers.

        :param observer_altitude: Altitude of the observer in meters.
        :type observer_altitude: float
        :return: Altitude in kilometers.
        :rtype: float
        """
        return observer_altitude / 1000.0

    def _calculate_a0_star(self, observer_altitude: int) -> float:
        r"""
        Calculate :math:`a_0^*` based on observer altitude.

        The formula used to calculate :math:`a_0^*` is:

        .. math::
            a_0^* = 0.4237 - 0.00821 \times (6 - A)^2

        :param observer_altitude: Altitude of the observer in meters.
        :type observer_altitude: float
        :return: :math:`a_0^*` value.
        :rtype: float
        """
        observer_altitude_km = self._convert_to_km(observer_altitude)
        observer_altitude_diff = 6.0 - observer_altitude_km
        return 0.4237 - 0.00821 * observer_altitude_diff**2

    def _calculate_a1_star(self, observer_altitude: int) -> float:
        r"""
        Calculate :math:`a_1^*` based on observer altitude.

        The formula used to calculate :math:`a_1^*` is:

        .. math::
            a_1^* = 0.5055 + 0.00595 \times (6.5 - A)^2

        :param observer_altitude: Altitude of the observer in meters.
        :type observer_altitude: float
        :return: :math:`a_1^*` value.
        :rtype: float
        """
        observer_altitude_km = self._convert_to_km(observer_altitude)
        observer_altitude_diff = 6.5 - observer_altitude_km
        return 0.5055 + 0.00595 * observer_altitude_diff**2

    def _calculate_k_star(self, observer_altitude: int) -> float:
        r"""
        Calculate :math:`k^*` based on observer altitude.

        The formula used to calculate :math:`k^*` is:

        .. math::
            k^* = 0.2711 + 0.01858 \times (2.5 - A)^2

        :param observer_altitude: Altitude of the observer in meters.
        :type observer_altitude: float
        :return: :math:`k^*` value.
        :rtype: float
        """
        observer_altitude_km = self._convert_to_km(observer_altitude)
        observer_altitude_diff = 2.5 - observer_altitude_km
        return 0.2711 + 0.01858 * observer_altitude_diff**2

    def calculate_transmittance_components(
        self, climate_type: str, observer_altitude: int
    ) -> Tuple[float, float, float]:
        r"""
        Calculate the components of clear-sky beam radiation transmittance
        :math:`a_0`, :math:`a_1`, and :math:`k`
        based on climate type and observer altitude.


        Correction factors adjust the clear-sky beam
        radiation transmittance components. According
        to the following formulas:

        .. math::
            a_0 = r_0 \times a_0^*

            a_1 = r_1 \times a_1^*

            k = r_k \times k^*

        :param climate_type: Climate type
               (i.e., ``TROPICAL``, ``MIDLATITUDE SUMMER``,
               ``SUBARCTIC SUMMER``, or ``MIDLATITUDE WINTER``).
        :type climate_type: str
        :param observer_altitude: Altitude of the observer in meters.
        :type observer_altitude: float
        :return: Components of clear-sky beam radiation transmittance
                (:math:`a_0`, :math:`a_1`, :math:`k`).
        :rtype: tuple of floats
        :raises ValueError: If an invalid climate type is provided.
        """
        if climate_type.upper() not in self.CLIMATE_CONSTANTS:
            raise ValueError("Invalid climate type")

        r0, r1, rk = self.CLIMATE_CONSTANTS[climate_type.upper()]

        a0_star = self._calculate_a0_star(observer_altitude)
        a1_star = self._calculate_a1_star(observer_altitude)
        k_star = self._calculate_k_star(observer_altitude)

        return r0 * a0_star, r1 * a1_star, rk * k_star



# Copyright 2023 Alireza Aghamohammadi

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Optional

from .sun_position import SunPosition


class Observer:
    def __init__(
        self,
        observer_latitude: Optional[float] = None,
        observer_longitude: Optional[float] = None,
    ):
        """
        To instantiate the ``Observer`` class, provide the following parameter.

        :param observer_latitude: The latitude of the observer in degrees (optional).
        :type observer_latitude: Optional[float]
        :param observer_longitude: The longitude of the observer in degrees (optional).
        :type observer_longitude: Optional[float]
        """
        self.observer_latitude = (
            math.radians(observer_latitude) if observer_latitude is not None else None
        )
        self.observer_longitude = (
            math.radians(observer_longitude) if observer_longitude is not None else None
        )
        self.sun_position = SunPosition()

    def calculate_zenith_angle(self, day_of_year: int, solar_time: float) -> float:
        r"""
        Calculate the zenith angle.

        The zenith angle is calculated using the formula:

        .. math::
            \cos(\theta_z) = \sin(\phi) \times \sin(\delta)
            + \cos(\phi) \times \cos(\delta) \times \cos(\omega)


        | - :math:`\theta_z` is the zenith angle
        | - :math:`\phi` is the latitude of the observer
        | - :math:`\delta` is the solar declination
        | - :math:`\omega` is the hour angle.

        :param day_of_year: The day of the year.
        :type day_of_year: int
        :param solar_time: The solar time in seconds.
        :type solar_time: float
        :return: The zenith angle in radians.
        :rtype: float
        """
        observer_latitude = self._validate_latitude()

        solar_declination = self.sun_position.solar_declination(day_of_year)
        hour_angle = self.sun_position.hour_angle(solar_time)
        return math.acos(
            math.sin(observer_latitude) * math.sin(solar_declination)
            + math.cos(observer_latitude)
            * math.cos(solar_declination)
            * math.cos(hour_angle)
        )

    def calculate_sunrise_sunset(self, day_of_year: int) -> tuple:
        r"""
        Calculate the hour angle at sunrise and sunset.

        The hour angle at sunrise and sunset is calculated using the formula:

        .. math::
            \cos(\omega) = -\tan(\phi) \times \tan(\delta)


        | - :math:`\omega` is the hour angle
        | - :math:`\phi` is the latitude of the observer
        | - :math:`\delta` is the solar declination.

        :param day_of_year: The day of the year.
        :type day_of_year: int
        :return: The hour angle at sunrise and sunset in radians.
        :rtype: tuple
        """
        observer_latitude = self._validate_latitude()

        solar_declination = self.sun_position.solar_declination(day_of_year)
        hour_angle = math.acos(
            -math.tan(observer_latitude) * math.tan(solar_declination)
        )

        sunrise = -hour_angle
        sunset = hour_angle

        return sunrise, sunset

    def _validate_latitude(self) -> float:
        if self.observer_latitude is None:
            raise ValueError("Observer latitude must be provided.")
        return self.observer_latitude
